// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import BigInt from 'big-integer';
import superagent from 'superagent';

export default class TSignClient {
    constructor(url, user) {
        this.url = url + '/api';
        this.user = user;
        this._id = null;
    }

    getId() {
        if (this._id != null)
            return Promise.resolve(this._id);

        return superagent.get(`${this.url}/storage`).then(res => this._id = new BigInt(res.text));
    }

    getKi(message) {
        var req = BigInt.isInstance(message)
            ? superagent.get(`${this.url}/sign/ki/${this.user}/${message}`)
            : superagent.post(`${this.url}/sign/ki/${this.user}`).send(JSON.stringify(message)).set('Content-Type', 'application/json');
        
        return req.then(res => res.text);
    }

    getAlpha(ks) {
        return superagent.post(`${this.url}/sign/alpha/${this.user}`).send(ks).set('Content-Type', 'application/json').then(res => res.body);
    }

    getDelta(beta, alphas) {
        return superagent.post(`${this.url}/sign/delta/${this.user}/${encodeBase64Url(beta)}`).send(alphas).set('Content-Type', 'application/json').then(res => res.body);
    }

    getRSi(sigma, deltas) {
        return superagent.post(`${this.url}/sign/rsi/${this.user}/${encodeBase64Url(sigma)}`).send(deltas).set('Content-Type', 'application/json').then(res => decodeSiFromArray(res.text));
    }
}

function encodeBase64Url(text) {
    return text.replace(/\=/g, "").replace(/\+/g, "-").replace(/\//g, "_");
}

function decodeSiFromArray(signature) {
    var array = Buffer.from(signature, 'base64');
    var r = BigInt.fromArray(Array.from(array.slice(0, 32)), 256, false);
    var si = BigInt.fromArray(Array.from(array.slice(32, 64)), 256, false);
    var xi = BigInt.fromArray(Array.from(array.slice(64, 96)), 256, false);
    var yi = BigInt.fromArray(Array.from(array.slice(96, 128)), 256, false);

    return {r, si, xi, yi};
}

function decodeSignatureFromArray(signature) {
    var array = Buffer.from(signature, 'base64');
    var r = BigInt.fromArray(Array.from(array.slice(4).slice(0, array[3])), 256, false);
    var s = BigInt.fromArray(Array.from(array.slice(array[3] + 6)), 256, false);

    return [r, s];
}